################################################################################
## Classes v0.21.0p2-01 2022-02-01
################################################################################

#
# Overview of how the mod works:
#
# 1. Game creates Event and Girl objects at launch based on files in progress mod\variables\
#     - event_list.rpy contains information for all events such as their triggers
#     - variables.rpy contains information for all girls and other misc vars
# 2. When the progress screen is opened, functions in update status.rpy are run
# 3. When the progress screen/hint tracker is opened:
#     1. ProgressMod.update_all() is run, causing 50 to 75 Event objects to
#        run Event.update_status(), updating their vars and generating a new event hint 
#

init python:

    # Class for information about the game's events
    class Event:

        import string

        def __init__(self, name, girl, chapter, type, reqs, miss_preq = None, reqs1 = None, reqs2 = None):
            self.name = name                                            # the event's variable name
            self.girl = girl                                            # the girl associated with the event (Girl object)
            self.event_number = None                                    # the event's number in the girl's event list
            self.chapter = chapter                                      # the chapter the event appears in
            self.type = type                                            # the type of event (i.e, how it is triggered)
            self.reqs = reqs                                            # requirements to trigger the event
            self.miss_preq = miss_preq                                  # the event's missable prerequisites, if any

            self.or_event = False                                       # does the event have an "or" in its requirements
            self.or_reqs = ""                                           # original set of requirments including "or"
            self.reqs1 = reqs1                                          # first set of requirements (used if the switch is true)
            self.reqs2 = reqs2                                          # second set of requirements (used if the switch is false)
            self.chapter_end = False
            self.chapter_end_reqs = ""

            self.triggers = dict()
            self.triggers["event_checks"] = []
            self.triggers["stat_checks"] = []
            self.triggers["desire_req"] = 0
            self.triggers["totaldays"] = 0
            self.triggers["day_restrictions"] = []
            self.triggers["day_restriction_text"] = ""

            self.ready = False                                          # is the event ready to fire?
            self.hint = ""                                              # the event hint text to be displayed
            self.light_hint = "N/A"
            self.attention = False                                      # whether to add (!) flag to event hint
            self.completed = False                                      # has the event been completed?
            self.missed = False                                         # has the event been permanently missed?
            self.blocked = True                                         # is something blocking this event's hint from the tracker?
            self.blocked_reason = ""                                    # why the event is blocked

            self.previous_event = "None"                                # previous event (when event is part of a chain)
            self.chain_processed = False                                # whether the event has been processed (if a chain event)

            if self.type == "light":
                parts = self.reqs.split(", ")
                self.light_hint = parts[1]
                self.reqs = parts[0]

            if self.type == "chapter_end":
                self.chapter_end_reqs = self.reqs
                Main.chapter_progress(self)
                self.chapter_end = True

            if " or " in self.reqs:
                self.or_handling()
                if self.or_choice() == 1:
                    self.reqs = self.reqs1
                else:
                    self.reqs = self.reqs2

            # adds the event to the relevent girl's lists
            girl.event_list.append(self)
            self.event_number = self.girl.event_list.index(self)
            if self.girl in girls_list:
                if not self.type in ["chain", "chain_lust", "chain_lust_adv", "firsthall"]:
                    if "date" in self.type:
                        temp_list = self.girl.name + ".date_event_list"
                    else:
                        temp_list = self.girl.name + "." + self.type + "_event_list"
                    eval(temp_list).append(self)
            elif self.girl in [Main, Dream, Light]:
                if not self.type == "chain":
                    if self.girl == Main:
                        temp_list = "Main." + self.type + "_event_list"
                    elif self.girl == Dream:
                        temp_list = "Dream." + self.type + "_event_list"
                    elif self.girl == Light:
                        temp_list = "Light." + self.type + "_event_list"
                    eval(temp_list).append(self)
            ProgressMod.event_list.append(self)

        # Method to break down an event's triggers (as found in self.reqs) by type

        def process_triggers(self):

            self.triggers["event_checks"] = []
            self.triggers["stat_checks"] = []
            self.triggers["desire_req"] = 0
            self.triggers["totaldays"] = 0
            self.triggers["day_restrictions"] = []
            self.triggers["day_restriction_text"] = ""

            trigger_list = self.reqs.split(" and ")
            for trigger in trigger_list:
                if trigger.split()[0] == self.name:
                    trigger_list.pop(trigger_list.index(trigger))
                elif trigger.split()[0] == "totaldays":
                    self.triggers["totaldays"] = int(trigger.split()[2])
                if self.girl.name.lower() + "_desire" in trigger:
                    self.triggers["desire_req"] = int(trigger.split()[2])

            for x in range(len(trigger_list)):
                current_trigger = trigger_list[x]
                if current_trigger[-4:] == "True":
                    current_name = current_trigger.split()[0]
                    for current_girl in girls_list:
                        for current_event in current_girl.event_list:
                            if current_name == current_event.name:
                                if current_name not in self.triggers["event_checks"]:
                                    self.triggers["event_checks"].append("ev_" + current_name)
                    for current_event in Main.event_list:
                        if current_name == current_event.name:
                            if current_name not in self.triggers["event_checks"]:
                                self.triggers["event_checks"].append("ev_" + current_name)
                    for current_event in Dream.event_list:
                        if current_name == current_event.name:
                            if current_name not in self.triggers["event_checks"]:
                                self.triggers["event_checks"].append("ev_" + current_name)
                    for current_event in Light.event_list:
                        if current_name == current_event.name:
                            if current_name not in self.triggers["event_checks"]:
                                self.triggers["event_checks"].append("ev_" + current_name)
                elif current_trigger[0:3] == "day":
                    self.triggers["day_restrictions"].append(current_trigger)
                elif "_desire" in current_trigger:
                    self.triggers["stat_checks"].append(current_trigger)
                elif "_compassion" in current_trigger:
                    self.triggers["stat_checks"].append(current_trigger)

            if len(self.triggers["day_restrictions"]) != 0:
                self.triggers["day_restrictions"].sort()
                self.triggers["day_restriction_text"] = self.generate_day_restrictions_text(*self.triggers["day_restrictions"])

        # Method to handle events which have an "or" in their trigger by breaking that down into
        # two sets of event requirements. Currently written around the only two events that fall
        # into this category. Will likely have to be adjusted for future events.

        def or_handling(self):

            self.or_event = True
            self.or_reqs = self.reqs
            parens = False          # do the event's triggers include parentheses?
            paren_start = 0         # the location of the first (
            paren_stop = 0          # the location of the final )

            if "(" in self.or_reqs and ")" in self.or_reqs:
                parens = True
                i = 0
                for c in self.or_reqs:
                    if c == "(" and paren_start == 0:
                        paren_start = i
                    if c == ")":
                        paren_stop = i
                    i = i + 1
                paren_substring = self.or_reqs[paren_start:paren_stop+1]    # "(rindate50 == True or (rindorm50special == True and rinbetrayed == True))"
                temp_reqs = self.or_reqs.replace(paren_substring,"")        # "totaldays >= 400 and secondbeach18 == True and and ramen30 == True and mollydorm30 == True and nikidate15 == True and day == 5 and halloweentwo1 == False"
                if " and  and " in temp_reqs:
                    temp_reqs = temp_reqs.replace(" and  and ", " and ")    # "totaldays >= 400 and secondbeach18 == True and ramen30 == True and mollydorm30 == True and nikidate15 == True and day == 5 and halloweentwo1 == False"
                or_place = paren_substring.find(" or ")
                or_first_half = paren_substring[0:or_place]           # "(rindate50 == True"
                or_second_half = paren_substring[or_place+4:]         # "(rindorm50special == True and rinbetrayed == True))"
                or_first_half = or_first_half.replace("(", "")        # "rindate50 == True"
                or_first_half = or_first_half.replace(")", "")
                or_second_half = or_second_half.replace("(", "")      # "rindorm50special == True and rinbetrayed == True))"
                or_second_half = or_second_half.replace(")", "")      # "rindorm50special == True and rinbetrayed == True"
                self.reqs1 = temp_reqs + " and " + or_first_half
                self.reqs2 = temp_reqs + " and " + or_second_half
            else:
                or_place = self.or_reqs.find(" or ")
                or_first_half = self.or_reqs[0:or_place]              # "totaldays >= 139 and chika_lust >= 5 and chikadorm20 == True and mall20 == True and chikadetention == True and day139 == False"
                or_second_half = self.or_reqs[or_place+4:]            # "chikadorm20 == True and mall20 == True and chika_lust >= 10 and day139 == False"
                self.reqs1 = or_first_half
                self.reqs2 = or_second_half

        # Method used to determine which set of requirements an event with an "or" in
        # its triggers should currently be using

        def or_choice(self):

            choose_first = False
            choose_second = False
            or_num = 1

            or_triggers1 = self.reqs1.split(" and ")
            or_triggers2 = self.reqs2.split(" and ")
            for t in or_triggers1:
                if t.split()[0] in ["kiobjevent1", "kiohandjob", "hanapervert", "hanadate1"]:
                    if eval(t) == False:
                        choose_second = True
                    else:
                        choose_first = True
            for t in or_triggers2:
                if t.split()[0] in ["kiobjevent1", "kiohandjob", "hanapervert", "hanadate1"]:
                    if eval(t) == False:
                        choose_first = True
                    else:
                        choose_second = True

            if choose_first:
                or_num = 1
            else:
                or_num = 2

            return or_num

        # Method to convert a chain event's triggers and type to those of a normal event
        # its requirements are made to point to the previous event in the chain
        # and it is inserted into the proper event list for that girl

        def gather_chain_reqs(self):

            if self.type == "chain" and self.chain_processed == False:
                self.previous_event = self.reqs
                self.reqs = self.previous_event + " == True and " + self.name + " == False"
                self.process_triggers()
                # if not self.girl in [Main, Dream]:
                #     for current_girl in girls_list:
                #         found = False
                #         for current_event in current_girl.event_list:
                #             if current_event.name == self.previous_event:
                #                 self.type = current_event.type
                #                 temp_list = self.girl.name + "." + current_event.type + "_event_list"
                #                 if current_event.type in ["date_morning", "date_afternoon", "date_night", "date_work"]:
                #                     temp_list = self.girl.name + ".date_event_list"
                #                 if not temp_list == self.girl.name + "." + "chain_event_list":
                #                     if len(eval(temp_list)) == 0:
                #                         eval(temp_list).append(self)
                #                     elif self.event_number > eval(temp_list)[-1].event_number:
                #                         eval(temp_list).append(self)
                #                     else:
                #                         for e in range(len(eval(temp_list))):
                #                             if eval(temp_list)[e].event_number > self.event_number:
                #                                 eval(temp_list).insert(e, self)
                #                     found = True
                #                     break
                #         if found:
                #             break
                #     for current_event in Main.event_list:
                #         if current_event.name == self.previous_event:
                #             self.type = current_event.type
                #             temp_list = self.girl.name + "." + current_event.type + "_event_list"
                #             if len(eval(temp_list)) == 0:
                #                 eval(temp_list).append(self)
                #             elif self.event_number > eval(temp_list)[-1].event_number:
                #                 eval(temp_list).append(self)
                #             else:
                #                 for e in range(len(eval(temp_list))):
                #                     if eval(temp_list)[e].event_number > self.event_number:
                #                         eval(temp_list).insert(e, self)
                # elif self.girl == Dream:
                #     for current_event in Dream.event_list:
                #         if current_event.name == self.previous_event:
                #             self.type = current_event.type
                #             if not self.type == "chain":
                #                 temp_list = "Dream." + current_event.type + "_event_list"
                #                 order_number = eval(temp_list).index(current_event)
                #                 eval(temp_list).insert(order_number+1, self)
                #                 break
                # else:
                #     found = False
                #     for current_event in Main.event_list:
                #         if current_event.name == self.previous_event:
                #             found = True
                #             self.type = current_event.type
                #             if not self.type == "chain":
                #                 temp_list = "Main." + current_event.type + "_event_list"
                #                 order_number = eval(temp_list).index(current_event)
                #                 eval(temp_list).insert(order_number+1, self)
                #                 break
                #     if found == False:
                #         for current_event in Dream.event_list:
                #             found = True
                #             self.type = current_event.type
                #             if not self.type == "chain":
                #                 temp_list = "Dream." + current_event.type + "_event_list"
                #                 order_number = eval(temp_list).index(current_event)
                #                 eval(temp_list).insert(order_number+1, self)
                #                 break


            self.chain_processed = True
            if self.reqs.split()[0] == "totaldays":
                self.triggers["totaldays"] = int(self.reqs.split()[2])

        # Method to update the event's various variables to reflect the player's progress
        # Events will only be updated if they are not from a previous chapter and have not been completed

        def update_status(self):
            if self.or_event:
                if self.or_choice() == 1:
                    self.reqs = self.reqs1
                else:
                    self.reqs = self.reqs2
                self.process_triggers()
            if self.chapter_end:
                Main.chapter_progress(self)
            if eval(self.name) == False:
                self.process_triggers()
                self.completed = False
                self.missed = False
                if self.type == "chain" and self.chain_processed == False:
                    self.gather_chain_reqs()
                if self.miss_preq is not None:
                    if eval(self.miss_preq) == True:
                        self.missed = True
                if self.chapter == current_chapter:
                    self.hint = ""
                    self.generate_hint()
            else:
                self.completed = True
                self.hint = ""
            if not self.hint == "" and not self.hint == "Event will trigger automatically.":
                self.girl.has_hint = True

        # Method to provide a single event hint from the event's requirements
        # Example output: "Visit Sana's dorm room on Wednesday."
        # Hint precedence = Affection > Lust > Events

        def generate_hint(self, block_override = False):

            import string

            failed_stat_checks = 0
            failed_event_checks = 0
            self.day_restriction = ""                                   # string to hold one day of the week restriction for the hint
            continue_parsing = True                                     # tells the parser whether to stop
            skip_processing = False
            self.hint = ""
            self.ready = False

            # 1. Initial Trigger Processing
            if not self.miss_preq == None:
                if eval(self.miss_preq) == True:
                    self.missed = True
                    skip_processing = True
                    continue_parsing = False
            if continue_parsing == True:
                for trigger in self.triggers["event_checks"]:
                    current_event = eval(trigger)
                    if current_event.completed == False:
                        failed_event_checks += 1
                        break
                for trigger in self.triggers["stat_checks"]:
                    if eval(trigger) == False:
                        failed_stat_checks += 1
                        if "_desire" in trigger:
                            if "janu_desire" in trigger:
                                self.hint = "Increase Janu's desire to " + trigger.split()[-1] + "."
                            else:
                                affection_girl = self.get_girl_from_stat(trigger)
                                self.hint = affection_girl.desire[:-1] + " to " + trigger.split()[-1] + " (currently " + str(affection_girl.current_desire) + ")."
                        elif "_compassion" in trigger:
                            self.hint = "Increase Janu's compassion to " + trigger.split()[-1] + "."
                        break

            # 2. Event Triggers
            if failed_event_checks > 0 and failed_stat_checks == 0 and skip_processing == False:
                self.hint = self.missing_events(*self.triggers["event_checks"])

            # 3. Event Hint (if all triggers have been met)
            if failed_event_checks == 0 and failed_stat_checks == 0 and skip_processing == False:
                self.ready = True
                if not self.girl in [Main, Dream, Light]:
                    self.generate_hint_text()
                elif self.girl == Dream:
                    if self.type == "dream":
                        self.hint = "Sleep on a weekend night."
                elif self.girl == Light:
                    self.hint = self.light_hint
                elif self.girl == Main:
                    if self.type == "dorm_room":
                        self.hint = self.girl.dorm_room
                    elif len(self.triggers["day_restrictions"]) == 1:
                        if self.type == "weekend_night":
                            self.hint = self.girl.wait_weekend_night
                        elif self.type == "sat_night":
                            self.hint = self.girl.wait_saturday_night
                        else:
                            self.girl.wait_set_day(self.triggers["day_restriction_text"])
                            self.hint = self.girl.wait_day
                    elif self.type == "common_room":
                        self.hint = "Visit the common room (weekend)."
                    else:
                        Main.wait_set_time(self)
            if self.attention:
                self.hint = self.hint[:-1] + " {color=#EF1A1A}(!){/color}"
            if self.girl == Main and self.ready and self.type == "chain":
                self.hint = "Automatic event chain in progress."
            elif self.type == "chain" and self.ready and eval(self.previous_event):
                self.hint = "Automatic event chain in progress."

            # 4. Event Blocking
            if block_override == False:
                self.event_blocking()
                if self.blocked or self.missed:
                    self.hint = ""
            else:
                return self.hint

        # Method to create a string indicating progress is needed in one or more event types
        # Example input: "ev_cafe20", "ev_soccer20", "ev_streets15" (strings)
        # Example output: "Rin, Miku progress needed."

        def missing_events(self, *event_list):

            missing_list = ""                                                                                   # string to hold the events that the player still needs to complete
            already_added = []                                                                                  # used to prevent duplicates in list of girls
            for event in event_list:                                                                       # for each event in the provided list,
                event_name = eval(event)
                if not event_name.completed:                                                                    # check to see if the event has been seen
                    if event_name in Main.event_list and "main" not in already_added:                      # main events
                        missing_list = missing_list + Main.colored_name + ", "
                        already_added.append("main")
                    if event_name in Dream.event_list and "dream" not in already_added:
                        missing_list = missing_list + Dream.colored_name + ", "
                        already_added.append("dream")
                    if event_name in Light.event_list and "light" not in already_added:
                        missing_list = missing_list + Light.colored_name + ", "
                        already_added.append("light")
                    else:                                                                                       # girl events
                        for current_girl in girls_list:
                            if event_name in current_girl.event_list and current_girl.name not in already_added:
                                missing_list = missing_list + current_girl.colored_name + ", "
                                already_added.append(current_girl.name)
                                break
            missing_list = missing_list[:-2] + " progress needed."
            return missing_list

        # Method to strip Boolean checks from a string
        # Example input: "kaoridate15p3 == True"
        # Example output: ev_kaoridate15p3 (Event object)
        def get_event(self, current_trigger):

            if current_trigger[-4:] == "True":
                current_trigger = current_trigger[:-8]
            elif current_trigger[-5:] == "False":
                current_trigger = current_trigger[:-9]
            current_trigger = eval("ev_" + current_trigger)
            return current_trigger

        # Method to determine the girl associated with a given affection or lust check
        # Example input: "touka_love >= 5"
        # Example output: Touka (Girl object)
        def get_girl_from_stat(self, var_name):

            component = var_name.split()                                 # splits the check into its component parts ("touka_love", ">=", "5")
            var_name = component[0]                                      # first component of the check ("touka_love")
            var_name = var_name.capitalize()                             # "Touka_love"
            var_name = var_name[:-7]                                     # "Touka"
            var_name = eval(var_name)                                    # Touka (Girl object)
            return var_name

        # Method to generate text explaining the day of the week restrictions for an event
        # Example input: "day != 2", "day < 5"
        # Example output: "a weekday other than Tuesday"
        def generate_day_restrictions_text(self, *temp_list):

            weekend_restriction = False                             # whether the event is weekend only
            weekday_restriction = False                             # whether the event is weekday only
            restriction_text = ""
            temp_restriction = ""
            exclude_day = []                                        # list of days the event cannot happen on
            restriction_list = []                                   # list of all restrictions
            greater_than = False
            lesser_than = False
            min_range = 0
            max_range = 7

            for element in temp_list:                               # copies input to new list (needed by Ren'py for some reason)
                restriction_list.append(element)

            for element in restriction_list:                                                                # goes through each element and
                if "!=" in element:                                                                         # adds any prohibited days to the appropriate list
                    exclude_day.append(element)
                if ">" in element:                                                                          # designates the event as weekend only
                    if "> 5" in element:
                        weekend_restriction = True
                    greater_than = True
                    min_range = int(element[-1:]) + 1                                                       # ("+1" is needed to exclude the first day from range)
                if "<" in element:                                                                          # designates the event as weekday only
                    if "< 6" in element:
                        weekday_restriction = True
                    lesser_than = True
                    max_range = int(element[-1:])

            if greater_than and lesser_than:                                                                # if there are both > and < in the restrictions
                weekday_restriction = False                                                                 # override other weekend/weekday designations
                weekend_restriction = False

            if len(exclude_day) >= 2:
                if "day != 6" in exclude_day and "day != 7" in exclude_day:                                 # replaces "day != 6 and day != 7" with "day < 6"
                    weekday_restriction = True
                    exclude_day.remove("day != 6")                   
                    exclude_day.remove("day != 7")
                    restriction_list.remove("day != 6")
                    restriction_list.remove("day != 7")
                    restriction_list.append("day < 6")
            if weekend_restriction == True and len(restriction_list) == 1:                                  # "Weekend"
                restriction_text = "Weekend"
            elif weekday_restriction == True and len(restriction_list) == 1:                                # "a weekday"
                restriction_text = "a weekday"
            elif "==" in restriction_list[0]:                                                               # "Tuesday"
                temp = restriction_list[0]
                restriction_text = days_of_the_week[int(temp[-1:])]
            else:
                if weekday_restriction == True:                                                             # "a weekday other than [Monday]"
                    restriction_text = "a weekday other than "
                elif greater_than and lesser_than:                                                              # "day > 2 and day < 6"
                    for i in range(min_range, max_range):
                        restriction_text = restriction_text + days_of_the_week[i] + " or "                      # "Wednesday or Thursday or Friday"
                    if len(range(min_range, max_range)) > 2:
                        temp_restriction = restriction_text.split()[0]
                        for i in range(len(range(min_range, max_range)) - 2):                                   # "- 2" to exclude the first and last days
                            temp_restriction = temp_restriction + ", " + restriction_text.split()[(i+1)*2]
                        temp_restriction = temp_restriction + ", or " + restriction_text.split()[-2] + " or "
                        restriction_text = temp_restriction                                                     # "Wednesday, Thursday, or Friday"
                else:                                                                                       # "not [Tuesday or Friday]"
                    restriction_text = restriction_text + "not "
                for i in exclude_day:
                    temp = i
                    restriction_text = restriction_text + days_of_the_week[int(temp[-1:])] + " or "
                restriction_text = restriction_text[:-4]
            if "date" in self.type or self.type == "sat_night":
                time = ""
                time = self.type.split("_")[1]
                if not self.type == "date_work":
                    if not restriction_text == "":
                        restriction_text = restriction_text + " " + time
            return restriction_text

        # Method to generate hint for an event whose triggers have all been met
        # Example output: "Visit Sana's dorm room'"
        def generate_hint_text(self):

            if self.type in ["work", "work2", "invite"]:
                temp_hint = self.girl.name + ".visit_" + self.type
                if self.type == "invite":
                    temp_hint = self.girl.name + "." + self.type
                temp_restrict = temp_hint + "_set_restriction('" + self.triggers["day_restriction_text"] + "')"
                if self.triggers["day_restriction_text"] == "":
                    self.hint = eval(temp_hint)
                else:
                    self.hint = eval(temp_restrict)
            elif self.type == "common_room":
                self.hint = self.girl.common_room
            elif self.type == "study_room":
                self.hint = "Invite " + self.girl.colored_name + " to the study room."
            elif self.type == "tutor":
                if ev_day58.completed:
                    self.hint = self.girl.tutor2
                else:
                    self.hint = self.girl.tutor
            elif "date" in self.type:
                time = ""
                time = self.type.split("_")[1]
                if not self.type == "date_work":
                    if self.triggers["day_restriction_text"] == "":
                        self.hint = eval(self.girl.name + ".call_" + time)
                    # else:
                    #     self.triggers["day_restriction_text"] = self.triggers["day_restriction_text"] + " " + time
                else:
                    self.hint = self.girl.visit_work
                if not self.triggers["day_restriction_text"] == "":
                    self.hint = self.girl.call_set_restriction(self.triggers["day_restriction_text"])
            elif self.type in ["weekday_morning", "weekend_morning", "weekend_afternoon", "weekend_night", "sat_morning", "sat_afternoon", "sat_night", "morning"]:
                if self.triggers["day_restriction_text"] == "":
                    if self.type == "sat_morning":
                        self.hint = self.girl.wait_saturday
                    elif self.type == "sat_afternoon":
                        self.hint = self.girl.wait_saturday_afternoon
                    elif self.type == "sat_night":
                        self.hint = self.girl.wait_saturday_night
                    elif self.type == "weekend_morning":
                        self.hint = self.girl.wait_weekend
                    elif self.type == "weekend_afternoon":
                        self.hint = self.girl.wait_weekend_afternoon
                    elif self.type == "weekend_night":
                        self.hint = self.girl.wait_weekend_night
                    elif self.type == "morning":
                        self.hint = "Wait for morning."
                    else:
                        self.hint = self.girl.wait_weekday
                else:
                    self.hint = self.girl.wait_set_restriction(self.triggers["day_restriction_text"])

        # Method to determine whether or not the hint for an incomplete event should appear in the appropriate event tracker
        # Events will be blocked from appearing if:
        #
        # 1. Its triggers include an unseen main event or an unseen event from the same girl
        # 2. The requisite number of days have not passed in the game
        # 3. The triggers include an event from an earlier chapter
        # 4a. For girls, there is an earlier unseen event of the same type
        # 4b. For main events, there is an ealier unseen event of the same type ready to fire
        # 5. Chain_lust events appear one update in advance to provide advance notice

        def event_blocking(self):

            self.blocked = False                                        # resets blocked flag

            # blocks the event if there are triggers for unseen main events or events from the same girl,
            # if one of the events being checked for has been blocked, or if an event requires a currently hidden girl

            for event in self.triggers["event_checks"]:
                trigger = eval(event)
                if trigger.completed:
                    trigger.blocked = False
                else:
                    if trigger.girl == self.girl or trigger.girl in [Main]:
                        self.blocked = True
                        self.blocked_reason = "Unseen main event/same girl event in triggers"
                        break
                if trigger.blocked:
                    self.blocked = True
                    self.blocked_reason = "Blocked event in triggers"
                    break
                if not trigger.girl == Main and not trigger.girl in [Dream, Light]:
                    if not trigger.girl.active:
                        self.blocked = True
                        self.blocked_reason = "Event for inactive girl in triggers"
                        break

            # blocks the event if there is an earlier unseen event of the same type

            if self.girl in girls_list:
                if not self.type in ["firsthall", "chain", "chain_lust", "chain_lust_adv"]:
                    if not "date" in self.type:
                        temp_list = self.girl.name + "." + self.type + "_event_list"
                    else:
                        temp_list = self.girl.name + ".date_event_list"
                    order_number = eval(temp_list).index(self)
                    if not order_number == 0:
                        if eval(temp_list)[order_number-1].completed == False and eval(temp_list)[order_number-1].missed == False:
                            self.blocked = True
                            self.blocked_reason = "Earlier unseen event of the same type"
            if self.ready:
                if self.girl == Main:
                    if not self.type == "chain":
                        temp_list = "Main." + self.type + "_event_list"
                        if len(eval(temp_list)) == 0:
                            order_number = 0
                        else:
                            order_number = eval(temp_list).index(self)
                        if not order_number == 0:
                            if eval(temp_list)[order_number-1].completed == False and eval(temp_list)[order_number-1].missed == False:
                                if eval(temp_list)[order_number-1].ready:
                                    self.blocked = True
                                    self.blocked_reason = "Earlier unseen event of the same type"

            if self.reqs == "chapter_end":
                if not self.girl.event_list[self.event_number - 1].completed:
                    self.blocked = True

            if totaldays < self.triggers["totaldays"]:
                self.blocked = True
                self.blocked_reason = "Not enought days spent in game"

            if self.chapter < current_chapter:
                self.blocked = True
                self.blocked_reason = "Event from an earlier chapter"

    # Class for each girl with events
    class Girl:

        def __init__(self, name, color, cr_day, tutor_day, work, eventprefix1, work2 = "None"):
            self.name = name                                    # the girl's name
            self.color = color                                  # the color associated with her
            self.cr_day = cr_day                                # common room/courtyard day
            self.tutor_day = tutor_day                          # tutoring day
            self.work = work                                    # the non-dorm location where she has events
            self.work2 = work2
            self.event_list = []                                # list for holding all events for the girl
            self.has_hint = False                               # has an event with a non-blank hint
            self.max = []                                       # list of total number of events the girl has in each chapter
            self.max.append(0)                                  # done to make the list index match directly with chapter numbers
            self.current_max = 0                                # number of total number of events the girl has in the current chapter
            self.active = False                                 # has the girl become active in the game yet

            self.highest_completed = 0
            self.current_desire = 0
            self.current_points = 0
            self.next_desire = 0
            self.desire_active = True

            # list of events of each type
            self.common_room_event_list = []
            self.dorm_room_event_list = []
            self.home_event_list = []
            self.work_event_list = []
            self.work2_event_list = []
            self.tutor_event_list = []
            self.date_event_list = []
            self.study_room_event_list = []

            self.morning_event_list = []
            self.weekday_morning_event_list = []
            self.weekend_night_event_list = []
            self.sat_morning_event_list = []
            self.sat_afternoon_event_list = []
            self.sat_night_event_list = []

            # Creates a string with the girl's name and the needed Renpy code to make it her color
            # (doing this in three steps is a workaround to prevent a Renpy error when starting the game)
            coloring1 = "{color=" + self.color
            coloring2 = "}"
            self.colored_name = coloring1 + coloring2 + self.name + "{/color}"

            # Strings to provide specific event hints
            if self.name in ["Chinaka", "Sora", "Juna", "Roni", "Zena"]:
                self.common_room = "Visit " + self.colored_name + " in the courtyard" + " (" + cr_day + ")" + "."
            else:
                self.common_room = "Visit " + self.colored_name + " in the common room" + " (" + cr_day + ")" + "."
            self.tutor = "Tutor " + self.colored_name + " (" + tutor_day + ")" + "."
            self.tutor2 = "Tutor " + self.colored_name + "."
            if self.work.split()[0] == "the":
                self.visit_work = "Visit " + self.work + "."
            else:
                self.visit_work = self.work
            if self.work2.split()[0] == "the":
                self.visit_work2 = "Visit " + self.work2 + "."
            else:
                self.visit_work2 = self.work2
            self.call_morning = "Call " + self.colored_name + " in the morning."
            self.call_afternoon = "Call " + self.colored_name + " in the afternoon."
            self.call_night = "Call " + self.colored_name + " at night."
            self.desire = "Increase " + self.colored_name + "'s desire."
            self.wait = "Wait for more time to pass."
            self.wait_weekday = "Wait for a weekday."
            self.wait_weekend = "Wait for the weekend."
            self.wait_weekend_afternoon = "Wait for the weekend (afternoon)"
            self.wait_weekend_night = "Wait for the weekend (night)."
            self.wait_saturday = "Wait for Saturday."
            self.wait_saturday_afternoon = "Wait for Saturday afternoon."
            self.wait_saturday_night = "Wait for Saturday night."
            self.missed = "You will not be able to see this event."

        # Methods to create strings incorporating additional requirements, i.e. "not Tuesday" or "a weekday"
        def call_set_restriction(self, restriction):
            call_restricted = "Call " + self.colored_name + " (" + restriction + ")."
            return call_restricted
        def wait_set_restriction(self, restriction):
            wait_restricted = "Wait until " + restriction + "."
            return wait_restricted

        # Method to create a string incorporating a time requirement, i.e. "not Tuesday" or "a weekday"
        def wait_set_day(self, restriction):
            self.wait_day = "Wait until " + restriction + "."

        # Method to update variables related to a girl's progress
        def progress_check(self):

            self.current_points = 0
            if eval(self.name.lower() + "_desire") == "N/A":
                self.current_desire = 0
            else:
                self.current_desire = eval(self.name.lower() + "_desire")

            for current_event in self.event_list:
                if current_event.completed:
                    self.current_points = self.current_points + 1

            event_progress = self.current_points - 1
            if event_progress < 0:
                event_progress = 0
            self.highest_completed = event_progress

            return event_progress

    # Class to hold methods and variables not specific to a particular girl
    class StoryEvent:

        def __init__(self, name, color):

            self.name = name
            self.color = color
            self.max = []                                       # total number of events the girl has in the current chapter
            self.max.append(0)
            self.current_max = 0
            self.has_hint = False                               # has an event with a non-blank hint
            self.points = 0
            self.highest_completed = 0
            self.active = True

            # Copied from Girl
            coloring1 = "{color=" + self.color
            coloring2 = "}"
            self.colored_name = coloring1 + coloring2 + self.name + "{/color}"

            self.event_list = []                                # list for holding all events for the girl

            # String to provide non-girl-specific event hints
            self.wait = "Wait for more time to pass."
            self.wait_weekday = "Wait for a weekday."
            self.wait_weekend = "Wait for the weekend."
            self.wait_weekend_afternoon = "Wait for the weekend (afternoon)"
            self.wait_weekend_night = "Wait for the weekend (night)."
            self.wait_saturday = "Wait for Saturday."
            self.wait_saturday_afternoon = "Wait for Saturday afternoon."
            self.wait_saturday_night = "Wait for Saturday night."
            self.missed = "You will not be able to see this event."

            # Variables that are set in the event trackers
            self.event_hint = ""        # used by parse_triggers(), stores the final event hint to be given to the player
            self.date_type = ""         # used by parse_triggers() to distinguish between date event types
            self.lust_type = ""         # used by parse_triggers() to distinguish between lust event types
            self.attention = False      # used by parse_triggers() to mark events needing (!)

            self.weekday_morning_event_list = []
            self.weekend_morning_event_list = []
            self.weekend_afternoon_event_list = []
            self.weekend_night_event_list = []
            self.sat_morning_event_list = []
            self.sat_afternoon_event_list = []
            self.sat_night_event_list = []
            self.common_room_event_list = []
            self.dorm_room_event_list = []
            self.dream_event_list = []
            self.light_event_list = []

            self.dorm_room = "Visit the dorms."

        # Method to create a string incorporating a time requirement, i.e. "not Tuesday" or "a weekday"
        def wait_set_day(self, restriction):
            self.wait_day = "Wait until " + restriction + "."

        # method to create a string including a generic time requirement
        def wait_set_time(self, target_event):
            if target_event.type == "sat_morning":
                target_event.hint = self.wait_saturday
            elif target_event.type == "sat_afternoon":
                target_event.hint = self.wait_saturday_afternoon
            elif target_event.type == "sat_night":
                target_event.hint = self.wait_saturday_night
            elif target_event.type == "weekend_morning":
                target_event.hint = self.wait_weekend
            elif target_event.type == "weekend_afternoon":
                target_event.hint = self.wait_weekend_afternoon
            elif target_event.type == "weekend_night":
                target_event.hint = self.wait_weekend_night
            else:
                target_event.hint = self.wait_weekday

# sat_morning, hanapervert : hanafinger1 / hanabc1 - kiohandjob : kiofinger1 / kiotutor2, totaldays >=77 and day76 == True and aniigym4 == True and diceevent1 == True and bookstorenirima3 == True and sorasleepover == True and christmas12 == False and day == 6
# hanapervert : hanafinger1 / hanabc1, kiohandjob : kiofinger1 / kiotutor2
# hanapervert : hanafinger1 / hanabc1
# kiohandjob : kiofinger1 / kiotutor2

# if hanapervert:
#     hanafinger1
# else:
#     hanabc1

# if kiohandjob:
#     kiofinger1
# else:
#     kiotutor2

        def chapter_progress(self, target_event):

            parts = target_event.chapter_end_reqs.split(", ")
            target_event.type = parts[0]
            target_event.reqs = parts[2]
            pairs = parts[1].split(" - ")
            for pair in pairs:
                switch = pair.split(": ")[0]
                condit_reqs = pair.split(": ")[1]
                if eval(switch):
                    target_event.reqs = target_event.reqs + " and " + condit_reqs.split(" / ")[0] + " == True"
                else:
                    target_event.reqs = target_event.reqs + " and " + condit_reqs.split(" / ")[1] + " == True"

        def progress_check(self):
            event_progress = "none"
            for current_event in self.event_list:
                current_event.completed = False
                if eval(current_event.name):
                    current_event.completed = True
                    self.highest_completed = current_event.event_number
                    if event_progress == "none":
                        event_progress = 0
                    else:
                        event_progress = event_progress + 1
            if event_progress == "none":
                event_progress = 0
            return event_progress

    # class to hold non-girl and event specific variables and methods

    class SOMod:

        def __init__(self):

            self.event_list = []

        # method to update the hints and certain variables for all events
        def update_all(self):

            globals()['current_chapter'] = 1
            if tochaptertwo:
                globals()['current_chapter'] = 2
            if tochapterthree:
                globals()['current_chapter'] = 3

            for current_girl in girls_list:
                current_girl.progress_check()
                if current_girl.active:
                    for current_event in current_girl.event_list:
                        current_event.hint = ""
                        if current_chapter == 1 and current_event.girl.highest_completed < 2:
                            if current_event.event_number < current_event.girl.highest_completed + 5:
                                current_event.update_status()
                        elif current_event.event_number < current_event.girl.highest_completed + 3:
                            current_event.update_status()
                        else:
                            if current_event.chapter <= current_chapter:
                                if eval(current_event.name) == False:
                                    current_event.completed = False
                                else:
                                    current_event.completed = True
            for current_event in Main.event_list:
                Main.progress_check()
                current_event.hint = ""
                current_event.update_status()
            for current_event in Dream.event_list:
                current_event.hint = ""
                current_event.update_status()
            for current_event in Light.event_list:
                current_event.hint = ""
                current_event.update_status()